@@ -17,7 +17,6 @@
 
 #include <grpc/support/port_platform.h>
 
-#include <type_traits>
 #include <utility>
 
 #include "absl/status/status.h"
@@ -58,8 +57,8 @@ template <typename Reader, typename Action>
 class ForEach {
  private:
   using ReaderNext = decltype(std::declval<Reader>().Next());
-  using ReaderResult = typename PollTraits<decltype(
-      std::declval<ReaderNext>()())>::Type::value_type;
+  using ReaderResult = typename PollTraits<
+      decltype(std::declval<ReaderNext>()())>::Type::value_type;
   using ActionFactory = promise_detail::PromiseFactory<ReaderResult, Action>;
   using ActionPromise = typename ActionFactory::Promise;
 
